// Transpose with and without shared plus single thread CPU
// Change N below to test different sizes

// Compile like this:

// cc -o transp transp-cs.c milli.c CSutilities.c Linux/MicroGlut.c -I Linux/ -DGL_GLEXT_PROTOTYPES -lm -lGL -lX11

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
//#include <GL/glut.h>
#include <GL/gl.h>
#include "MicroGlut.h"
#include "CSutilities.h"
#include "milli.h"

//#define N 16

// Using two shaders.
// 16 = CPU wins big, 43 (naive 55) to 4
// 64 = GPU barely wins, 48 (naive 64) to 71
// 128 = GPU wins 53 (naive 62) to 372
// 256 = GPU wins 74 (naive 90) to 1705 (2224?)
// 512 = GPU wins 86 (naive 108) to 13758

#define runs 5000

float *ac;
float *c;
int N = 64; // default, change with command-line argument

void transposeCPU()
{
	for (int x = 0; x < N; x++)
	for (int y = 0; y < N; y++)
		c[x + y*N] = ac[x*N + y];
}

int main(int argc, char **argv)
{
// Let MicroGLUT create a GL context
	glutInit (&argc, argv);
	glutInitContextVersion(4, 5);
	glutCreateWindow("Transp");
		
	if (argc >= 2)
	{
		N = atoi(argv[1]);
	}
	
	ac = (float *)calloc(N, N*sizeof(float));
	c = (float *)calloc(N, N*sizeof(float));
//	ac = (float *)malloc(N*N*sizeof(float));
//	c = (float *)malloc(N*N*sizeof(float));
   
	printf ("Vendor: %s\n", glGetString (GL_VENDOR));
	printf ("Renderer: %s\n", glGetString (GL_RENDERER));
	printf ("Version: %s\n", glGetString (GL_VERSION));
	printf ("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));


// Load and compile the compute shader    
    GLuint naive = loadComputeShader("transp.cs"); // Naive
    GLuint shared = loadComputeShader("transp2.cs"); // Shared - loses?!

	GLuint ssbo, ssbo3; //Shader Storage Buffer Object

	// Some data
	int i;
	for (i = 0; i < N*N; i++) ac[i]=i;

	printf("allocating on GPU\n");

	glUseProgram(naive);

// Create buffer, upload data
	glGenBuffers(1, &ssbo);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo);
	glBufferData(GL_SHADER_STORAGE_BUFFER, N*N * sizeof(float), ac, GL_STATIC_DRAW);
	
// Tell it where the input goes!
// The "5" matches a "layout" number in the shader.
// (Can we ask the shader about the number? I must try that.)
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 5, ssbo);

// Same for the other buffer, offsets, ID 7
	glGenBuffers(1, &ssbo3);
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo3);
	glBufferData(GL_SHADER_STORAGE_BUFFER, N*N * sizeof(float), c, GL_STATIC_DRAW);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 7, ssbo3);


// We must tell the other shader of the locations.
	glUseProgram(shared);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 5, ssbo);
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 7, ssbo3);

	printf("Running on CPU\n");

// Test the CPU
	ResetMilli();

	for (int i = 0; i < runs; i++)
		transposeCPU();

	float cpuTime = (float)GetMicroseconds()/1000.0;
	printf("CPU time:     %0.3f ms\n", cpuTime);

// Naive GPU
	ResetMilli();

	glUseProgram(naive);
	for (int i = 0; i < runs; i++)
	{
	    glDispatchCompute(N/16, N/16, 1); //Work groups launch
	    glFlush();
	}

	float naiveTime = (float)GetMicroseconds()/1000.0;
	printf("GPU Time (naive):     %0.3f ms\n", naiveTime);

// GPU using shared memory for coalescing
	ResetMilli();

	glUseProgram(shared);
	for (int i = 0; i < runs; i++)
	{
	    glDispatchCompute(N/16, N/16, 1); //Work groups launch
	    glFlush();
	}

	float sharedTime = (float)GetMicroseconds()/1000.0;
	printf("GPU Time (shared):     %0.3f ms\n", sharedTime);


// Get data back!
	glBindBuffer(GL_SHADER_STORAGE_BUFFER, ssbo3);
	float *ptr = (float *)glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_READ_ONLY);
// Convert int to string:
	if (N < 64)
	for (int y=0; y < N; y++)
	{
		for (int x=0; x < N; x++)
		{
			printf("%.1f ", ptr[N*y+x]);
		}
		printf("\n");
	}
}

